@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2014, 2016, 2018,
+ * Copyright (C) 1986, 1988, 1989, 1991-2014, 2016, 2018, 2019, 2020,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -35,12 +35,13 @@ static size_t SUBSEPlen;
 static char *SUBSEP;
 static char indent_char[] = "    ";
 
+static int sort_up_value_type(const void *p1, const void *p2);
 static NODE **null_lookup(NODE *symbol, NODE *subs);
 static NODE **null_dump(NODE *symbol, NODE *subs);
-static afunc_t null_array_func[] = {
+static const array_funcs_t null_array_func = {
+	"null",
 	(afunc_t) 0,
 	(afunc_t) 0,
-	null_length,
 	null_lookup,
 	null_afunc,
 	null_afunc,
@@ -53,23 +54,20 @@ static afunc_t null_array_func[] = {
 
 #define MAX_ATYPE 10
 
-static afunc_t *array_types[MAX_ATYPE];
+static const array_funcs_t *array_types[MAX_ATYPE];
 static int num_array_types = 0;
 
-/* array func to index mapping */
-#define AFUNC(F) (F ## _ind)
-
 /* register_array_func --- add routines to handle arrays */
 
-int
-register_array_func(afunc_t *afunc)
+static int
+register_array_func(const array_funcs_t *afunc)
 {
 	if (afunc && num_array_types < MAX_ATYPE) {
-		if (afunc != str_array_func && ! afunc[AFUNC(atypeof)])
+		if (afunc != & str_array_func && afunc->type_of == NULL)
 			return false;
 		array_types[num_array_types++] = afunc;
-		if (afunc[AFUNC(ainit)])	/* execute init routine if any */
-			(void) (*afunc[AFUNC(ainit)])(NULL, NULL);
+		if (afunc->init)	/* execute init routine if any */
+			(void) (*afunc->init)(NULL, NULL);
 		return true;
 	}
 	return false;
@@ -81,10 +79,10 @@ register_array_func(afunc_t *afunc)
 void
 array_init()
 {
-	(void) register_array_func(str_array_func);	/* the default */
+	(void) register_array_func(& str_array_func);	/* the default */
 	if (! do_mpfr) {
-		(void) register_array_func(int_array_func);
-		(void) register_array_func(cint_array_func);
+		(void) register_array_func(& int_array_func);
+		(void) register_array_func(& cint_array_func);
 	}
 }
 
@@ -98,7 +96,7 @@ make_array()
 	getnode(array);
 	memset(array, '\0', sizeof(NODE));
 	array->type = Node_var_array;
-	array->array_funcs = null_array_func;
+	array->array_funcs = & null_array_func;
 	/* vname, flags, and parent_array not set here */
 
 	return array;
@@ -111,7 +109,7 @@ void
 null_array(NODE *symbol)
 {
 	symbol->type = Node_var_array;
-	symbol->array_funcs = null_array_func;
+	symbol->array_funcs = & null_array_func;
 	symbol->buckets = NULL;
 	symbol->table_size = symbol->array_size = 0;
 	symbol->array_capacity = 0;
@@ -129,7 +127,7 @@ static NODE **
 null_lookup(NODE *symbol, NODE *subs)
 {
 	int i;
-	afunc_t *afunc = NULL;
+	const array_funcs_t *afunc = NULL;
 
 	assert(symbol->table_size == 0);
 
@@ -139,7 +137,7 @@ null_lookup(NODE *symbol, NODE *subs)
 	 */
 	for (i = num_array_types - 1; i >= 1; i--) {
 		afunc = array_types[i];
-		if (afunc[AFUNC(atypeof)](symbol, subs) != NULL)
+		if (afunc->type_of(symbol, subs) != NULL)
 			break;
 	}
 	if (i == 0 || afunc == NULL)
@@ -150,15 +148,6 @@ null_lookup(NODE *symbol, NODE *subs)
 	return symbol->alookup(symbol, subs);
 }
 
-/* null_length --- default function for array length interface */
-
-NODE **
-null_length(NODE *symbol, NODE *subs ATTRIBUTE_UNUSED)
-{
-	static NODE *tmp;
-	tmp = symbol;
-	return & tmp;
-}
 
 /* null_afunc --- default function for array interface */
 
@@ -618,6 +607,9 @@ do_delete(NODE *symbol, int nsubs)
 
 	(void) assoc_remove(symbol, subs);
 	DEREF(subs);
+	if (assoc_empty(symbol))
+		/* last element was removed, so reset array type to null */
+		null_array(symbol);
 
 #undef free_subs
 }
@@ -788,7 +780,7 @@ do_adump(int nargs)
 	}
 	symbol = POP_PARAM();
 	if (symbol->type != Node_var_array)
-		fatal(_("adump: first argument not an array"));
+		fatal(_("%s: first argument is not an array"), "adump");
 
 	ndump.type = Node_dump_array;
 	ndump.adepth = depth;
@@ -805,7 +797,7 @@ asort_actual(int nargs, sort_context_t ctxt)
 {
 	NODE *array, *dest = NULL, *result;
 	NODE *r, *subs, *s;
-	NODE **list = NULL, **ptr, **lhs;
+	NODE **list = NULL, **ptr;
 	unsigned long num_elems, i;
 	const char *sort_str;
 	char save;
@@ -829,31 +821,33 @@ asort_actual(int nargs, sort_context_t ctxt)
 	if (nargs >= 2) {  /* 2nd optional arg */
 		dest = POP_PARAM();
 		if (dest->type != Node_var_array) {
-			fatal(ctxt == ASORT ?
-				_("asort: second argument not an array") :
-				_("asorti: second argument not an array"));
+			fatal(_("%s: second argument is not an array"),
+				ctxt == ASORT ? "asort" : "asorti");
 		}
 	}
 
 	array = POP_PARAM();
 	if (array->type != Node_var_array) {
-		fatal(ctxt == ASORT ?
-			_("asort: first argument not an array") :
-			_("asorti: first argument not an array"));
+		fatal(_("%s: first argument is not an array"),
+			ctxt == ASORT ? "asort" : "asorti");
 	}
+	else if (array == symbol_table)
+		fatal(_("%s: first argument cannot be SYMTAB"),
+			ctxt == ASORT ? "asort" : "asorti");
+	else if (array == func_table)
+		fatal(_("%s: first argument cannot be FUNCTAB"),
+			ctxt == ASORT ? "asort" : "asorti");
 
 	if (dest != NULL) {
 		for (r = dest->parent_array; r != NULL; r = r->parent_array) {
 			if (r == array)
-				fatal(ctxt == ASORT ?
-					_("asort: cannot use a subarray of first arg for second arg") :
-					_("asorti: cannot use a subarray of first arg for second arg"));
+				fatal(_("%s: cannot use a subarray of first argument for second argument"),
+					ctxt == ASORT ? "asort" : "asorti");
 		}
 		for (r = array->parent_array; r != NULL; r = r->parent_array) {
 			if (r == dest)
-				fatal(ctxt == ASORT ?
-					_("asort: cannot use a subarray of second arg for first arg") :
-					_("asorti: cannot use a subarray of second arg for first arg"));
+				fatal(_("%s: cannot use a subarray of second argument for first argument"),
+					ctxt == ASORT ? "asort" : "asorti");
 		}
 	}
 
@@ -893,12 +887,7 @@ asort_actual(int nargs, sort_context_t ctxt)
 
 		for (i = 1, ptr = list; i <= num_elems; i++, ptr += 2) {
 			subs = make_number(i);
-			lhs = assoc_lookup(result, subs);
-			unref(*lhs);
-			*lhs = *ptr;
-			if (result->astore != NULL)
-				(*result->astore)(result, subs);
-			unref(subs);
+			assoc_set(result, subs, *ptr);
 		}
 	} else {
 		/* We want the values of the source array. */
@@ -913,11 +902,11 @@ asort_actual(int nargs, sort_context_t ctxt)
 			/* value node */
 			r = *ptr++;
 
-			if (r->type == Node_val) {
-				lhs = assoc_lookup(result, subs);
-				unref(*lhs);
-				*lhs = dupnode(r);
-			} else {
+			NODE *value;
+
+			if (r->type == Node_val)
+				value = dupnode(r);
+			else {
 				NODE *arr;
 				arr = make_array();
 				subs = force_string(subs);
@@ -926,13 +915,10 @@ asort_actual(int nargs, sort_context_t ctxt)
 				subs->stptr = NULL;
 				subs->flags &= ~STRCUR;
 				arr->parent_array = array; /* actual parent, not the temporary one. */
-				lhs = assoc_lookup(result, subs);
-				unref(*lhs);
-				*lhs = assoc_copy(r, arr);
+
+				value = assoc_copy(r, arr);
 			}
-			if (result->astore != NULL)
-				(*result->astore)(result, subs);
-			unref(subs);
+			assoc_set(result, subs, value);
 		}
 	}
 
@@ -979,7 +965,6 @@ cmp_strings(const NODE *n1, const NODE *n2)
 	char *s1, *s2;
 	size_t len1, len2;
 	int ret;
-	size_t lmin;
 
 	s1 = n1->stptr;
 	len1 = n1->stlen;
@@ -992,7 +977,9 @@ cmp_strings(const NODE *n1, const NODE *n2)
 		return 1;
 
 	/* len1 > 0 && len2 > 0 */
-	lmin = len1 < len2 ? len1 : len2;
+	// make const to ensure it doesn't change if we
+	// need to call memcmp(), below
+	const size_t lmin = len1 < len2 ? len1 : len2;
 
 	if (IGNORECASE) {
 		const unsigned char *cp1 = (const unsigned char *) s1;
@@ -1002,7 +989,9 @@ cmp_strings(const NODE *n1, const NODE *n2)
 			ret = strncasecmpmbs((const unsigned char *) cp1,
 					     (const unsigned char *) cp2, lmin);
 		} else {
-			for (ret = 0; lmin-- > 0 && ret == 0; cp1++, cp2++)
+			size_t count = lmin;
+
+			for (ret = 0; count-- > 0 && ret == 0; cp1++, cp2++)
 				ret = casetable[*cp1] - casetable[*cp2];
 		}
 		if (ret != 0)
@@ -1087,19 +1076,19 @@ static int
 sort_up_value_string(const void *p1, const void *p2)
 {
 	const NODE *t1, *t2;
+	int ret;
 
 	t1 = *((const NODE *const *) p1 + 1);
 	t2 = *((const NODE *const *) p2 + 1);
 
-	if (t1->type == Node_var_array) {
-		/* return 0 if t2 is a sub-array too, else return 1 */
-		return (t2->type != Node_var_array);
-	}
-	if (t2->type == Node_var_array)
-		return -1;		/* t1 (scalar) < t2 (sub-array) */
+	if (t1->type != Node_val || t2->type != Node_val)
+		return sort_up_value_type(p1, p2);
 
 	/* t1 and t2 both have string values */
-	return cmp_strings(t1, t2);
+	ret = cmp_strings(t1, t2);
+	if (ret != 0)
+		return ret;
+	return sort_up_index_string(p1, p2);
 }
 
 
@@ -1123,12 +1112,8 @@ sort_up_value_number(const void *p1, const void *p2)
 	t1 = *((NODE *const *) p1 + 1);
 	t2 = *((NODE *const *) p2 + 1);
 
-	if (t1->type == Node_var_array) {
-		/* return 0 if t2 is a sub-array too, else return 1 */
-		return (t2->type != Node_var_array);
-	}
-	if (t2->type == Node_var_array)
-		return -1;		/* t1 (scalar) < t2 (sub-array) */
+	if (t1->type != Node_val || t2->type != Node_val)
+		return sort_up_value_type(p1, p2);
 
 	ret = cmp_numbers(t1, t2);
 	if (ret != 0)
@@ -1138,9 +1123,10 @@ sort_up_value_number(const void *p1, const void *p2)
 	 * Use string value to guarantee same sort order on all
 	 * versions of qsort().
 	 */
-	t1 = force_string(t1);
-	t2 = force_string(t2);
-	return cmp_strings(t1, t2);
+	ret = cmp_strings(force_string(t1), force_string(t2));
+	if (ret != 0)
+		return ret;
+	return sort_up_index_string(p1, p2);
 }
 
 
@@ -1153,24 +1139,58 @@ sort_down_value_number(const void *p1, const void *p2)
 }
 
 
-/* sort_up_value_type --- qsort comparison function; ascending value type */
+/* do_sort_up_value_type --- backend comparison on ascending value type */
 
 static int
-sort_up_value_type(const void *p1, const void *p2)
+do_sort_up_value_type(const void *p1, const void *p2)
 {
 	NODE *n1, *n2;
 
+	static const NODETYPE element_types[] = {
+		Node_builtin_func,
+		Node_func,
+		Node_ext_func,
+		Node_var_new,
+		Node_var,
+		Node_var_array,
+		Node_val,
+		Node_illegal
+	};
+
 	/* we want to compare the element values */
 	n1 = *((NODE *const *) p1 + 1);
 	n2 = *((NODE *const *) p2 + 1);
 
-	/* 1. Arrays vs. scalar, scalar is less than array */
+	if (n1->type == Node_var && n2->type == Node_var) {
+		/* compare the values of the variables */
+		n1 = n1->var_value;
+		n2 = n2->var_value;
+	}
+
+	/* 1. Arrays vs. everything else, everything else is less than array */
 	if (n1->type == Node_var_array) {
 		/* return 0 if n2 is a sub-array too, else return 1 */
 		return (n2->type != Node_var_array);
 	}
 	if (n2->type == Node_var_array) {
-		return -1;		/* n1 (scalar) < n2 (sub-array) */
+		return -1;              /* n1 (non-array) < n2 (sub-array) */
+	}
+
+	/* 2. Non scalars */
+	if (n1->type != Node_val || n2->type != Node_val) {
+		int n1_pos, n2_pos, i;
+
+		n1_pos = n2_pos = -1;
+		for (i = 0; element_types[i] != Node_illegal; i++) {
+			if (n1->type == element_types[i])
+				n1_pos = i;
+
+			if (n2->type == element_types[i])
+				n2_pos = i;
+		}
+
+		assert(n1_pos != -1 && n2_pos != -1);
+		return (n1_pos - n2_pos);
 	}
 
 	/* two scalars */
@@ -1192,6 +1212,17 @@ sort_up_value_type(const void *p1, const void *p2)
 	return cmp_strings(n1, n2);
 }
 
+/* sort_up_value_type --- qsort comparison function; ascending value type */
+
+static int
+sort_up_value_type(const void *p1, const void *p2)
+{
+	int rc = do_sort_up_value_type(p1, p2);
+
+	/* use a tie-breaker if do_sort_up_value_type has no opinion */
+	return rc ? rc : sort_up_index_string(p1, p2);
+}
+
 /* sort_down_value_type --- qsort comparison function; descending value type */
 
 static int
@@ -1302,7 +1333,7 @@ assoc_list(NODE *symbol, const char *sort_str, sort_context_t sort_ctxt)
 		cmp_func = sort_funcs[qi].comp_func;
 		assoc_kind = sort_funcs[qi].kind;
 
-		if (symbol->array_funcs != cint_array_func)
+		if (symbol->array_funcs != & cint_array_func)
 			assoc_kind &= ~(AASC|ADESC);
 
 		if (sort_ctxt != SORTED_IN || (assoc_kind & AVALUE) != 0) {
