@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2018 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2020 the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -74,6 +74,9 @@ static bool resave_fs;
 static NODE *save_FS;		/* save current value of FS when line is read,
 				 * to be used in deferred parsing
 				 */
+static NODE *save_FPAT;		/* save current value of FPAT when line is read,
+				 * to be used in deferred parsing
+				 */
 static awk_fieldwidth_info_t *FIELDWIDTHS = NULL;
 
 NODE **fields_arr;		/* array of pointers to the field nodes */
@@ -94,16 +97,13 @@ init_fields()
 {
 	emalloc(fields_arr, NODE **, sizeof(NODE *), "init_fields");
 
-	getnode(fields_arr[0]);
-	*fields_arr[0] = *Nnull_string;
+	fields_arr[0] = make_string("", 0);
 	fields_arr[0]->flags |= NULL_FIELD;
 
 	parse_extent = fields_arr[0]->stptr;
 	save_FS = dupnode(FS_node->var_value);
 
-	getnode(Null_field);
-	*Null_field = *Nnull_string;
-	Null_field->valref = 1;
+	Null_field = make_string("", 0);
 	Null_field->flags = (STRCUR|STRING|NULL_FIELD); /* do not set MALLOC */
 
 	field0_valid = true;
@@ -288,7 +288,9 @@ set_record(const char *buf, int cnt, const awk_fieldwidth_info_t *fw)
 		memset(databuf, '\0', databuf_size);
 	}
 	/* copy the data */
-	memcpy(databuf, buf, cnt);
+	if (cnt != 0) {
+		memcpy(databuf, buf, cnt);
+	}
 
 	/*
 	 * Add terminating '\0' so that C library routines
@@ -338,6 +340,11 @@ reset_record()
 {
 	fields_arr[0] = force_string(fields_arr[0]);
 	purge_record();
+	if (api_parser_override) {
+		api_parser_override = false;
+		parse_field = normal_parse_field;
+		update_PROCINFO_str("FS", current_field_sep_str());
+	}
 }
 
 static void
@@ -391,6 +398,13 @@ set_NF()
 	nf = get_number_si(NF_node->var_value);
 	if (nf < 0)
 		fatal(_("NF set to negative value"));
+
+	static bool warned = false;
+	if (do_lint && NF > nf && ! warned) {
+		warned = true;
+		lintwarn(_("decrementing NF is not portable to many awk versions"));
+	}
+
 	NF = nf;
 
 	if (NF > nf_high_water)
@@ -452,7 +466,9 @@ re_parse_field(long up_to,	/* parse only up to this field number */
 	if (len == 0)
 		return nf;
 
-	if (RS_is_null && default_FS) {
+	bool default_field_splitting = (RS_is_null && default_FS);
+
+	if (default_field_splitting) {
 		sep = scan;
 		while (scan < end && (*scan == ' ' || *scan == '\t' || *scan == '\n'))
 			scan++;
@@ -493,7 +509,7 @@ re_parse_field(long up_to,	/* parse only up to this field number */
            			(long) (REEND(rp, scan) - RESTART(rp, scan)), sep_arr);
 		scan += REEND(rp, scan);
 		field = scan;
-		if (scan == end)	/* FS at end of record */
+		if (scan == end && ! default_field_splitting)	/* FS at end of record */
 			(*set)(++nf, field, 0L, n);
 	}
 	if (nf != up_to && scan < end) {
@@ -827,6 +843,16 @@ NODE **
 get_field(long requested, Func_ptr *assign)
 {
 	bool in_middle = false;
+	static bool warned = false;
+	extern int currule;
+	NODE *saved_fs;
+	Regexp *fs_regexp;
+
+	if (do_lint && currule == END && ! warned) {
+		warned = true;
+		lintwarn(_("accessing fields from an END rule may not be portable"));
+	}
+
 	/*
 	 * if requesting whole line but some other field has been altered,
 	 * then the whole line must be rebuilt
@@ -835,10 +861,18 @@ get_field(long requested, Func_ptr *assign)
 		if (! field0_valid) {
 			/* first, parse remainder of input record */
 			if (NF == -1) {
+				in_middle = (parse_high_water != 0);
+				if (current_field_sep() == Using_FPAT) {
+					saved_fs = save_FPAT;
+					fs_regexp = FPAT_regexp;
+				} else {
+					saved_fs = save_FS;
+					fs_regexp = FS_regexp;
+				}
 				NF = (*parse_field)(UNLIMITED - 1, &parse_extent,
 		    			fields_arr[0]->stlen -
 					(parse_extent - fields_arr[0]->stptr),
-		    			save_FS, FS_regexp, set_field,
+					saved_fs, fs_regexp, set_field,
 					(NODE *) NULL,
 					(NODE *) NULL,
 					in_middle);
@@ -895,17 +929,6 @@ get_field(long requested, Func_ptr *assign)
 		 */
 		if (parse_extent == fields_arr[0]->stptr + fields_arr[0]->stlen)
 			NF = parse_high_water;
-		else if (parse_field == fpat_parse_field) {
-			/* FPAT parsing is weird, isolate the special cases */
-			char *rec_start = fields_arr[0]->stptr;
-			char *rec_end = fields_arr[0]->stptr + fields_arr[0]->stlen;
-
-			if (    parse_extent > rec_end
-			    || (parse_extent > rec_start && parse_extent < rec_end && requested == UNLIMITED-1))
-				NF = parse_high_water;
-			else if (parse_extent == rec_start) /* could be no match for FPAT */
-				NF = 0;
-		}
 		if (requested == UNLIMITED - 1)	/* UNLIMITED-1 means set NF */
 			requested = parse_high_water;
 	}
@@ -929,18 +952,12 @@ static void
 set_element(long num, char *s, long len, NODE *n)
 {
 	NODE *it;
-	NODE **lhs;
 	NODE *sub;
 
 	it = make_string(s, len);
 	it->flags |= USER_INPUT;
 	sub = make_number((AWKNUM) (num));
-	lhs = assoc_lookup(n, sub);
-	unref(*lhs);
-	*lhs = it;
-        if (n->astore != NULL)
-                (*n->astore)(n, sub);
-	unref(sub);
+	assoc_set(n, sub, it);
 }
 
 /* do_split --- implement split(), semantics are same as for field splitting */
@@ -963,7 +980,7 @@ do_split(int nargs)
 		sep_arr = POP_PARAM();
 		if (sep_arr->type != Node_var_array)
 			fatal(_("split: fourth argument is not an array"));
-		if ((do_lint || do_lint_old) && ! warned) {
+		if ((do_lint_extensions || do_lint_old) && ! warned) {
 			warned = true;
 			lintwarn(_("split: fourth argument is a gawk extension"));
 		}
@@ -1018,7 +1035,7 @@ do_split(int nargs)
 
 			if (do_lint && ! warned) {
 				warned = true;
-				lintwarn(_("split: null string for third arg is a gawk extension"));
+				lintwarn(_("split: null string for third arg is a non-standard extension"));
 			}
 		} else if (fs->stlen == 1 && (sep->re_flags & CONSTANT) == 0) {
 			if (fs->stptr[0] == ' ') {
@@ -1129,7 +1146,7 @@ set_FIELDWIDTHS()
 	bool fatal_error = false;
 	NODE *tmp;
 
-	if (do_lint && ! warned) {
+	if (do_lint_extensions && ! warned) {
 		warned = true;
 		lintwarn(_("`FIELDWIDTHS' is a gawk extension"));
 	}
@@ -1292,13 +1309,13 @@ choose_fs_function:
 
 		set_parser(null_parse_field);
 
-		if (do_lint && ! warned) {
+		if (do_lint_extensions && ! warned) {
 			warned = true;
 			lintwarn(_("null string for `FS' is a gawk extension"));
 		}
-	} else if (fs->stlen > 1) {
+	} else if (fs->stlen > 1 || (fs->flags & REGEX) != 0) {
 		if (do_lint_old)
-			warning(_("old awk does not support regexps as value of `FS'"));
+			lintwarn(_("old awk does not support regexps as value of `FS'"));
 		set_parser(re_parse_field);
 	} else if (RS_is_null) {
 		/* we know that fs->stlen <= 1 */
@@ -1392,16 +1409,12 @@ current_field_sep_str()
 void
 update_PROCINFO_str(const char *subscript, const char *str)
 {
-	NODE **aptr;
 	NODE *tmp;
 
 	if (PROCINFO_node == NULL)
 		return;
 	tmp = make_string(subscript, strlen(subscript));
-	aptr = assoc_lookup(PROCINFO_node, tmp);
-	unref(tmp);
-	unref(*aptr);
-	*aptr = make_string(str, strlen(str));
+	assoc_set(PROCINFO_node, tmp, make_string(str, strlen(str)));
 }
 
 /* update_PROCINFO_num --- update PROCINFO[sub] with numeric value */
@@ -1409,16 +1422,12 @@ update_PROCINFO_str(const char *subscript, const char *str)
 void
 update_PROCINFO_num(const char *subscript, AWKNUM val)
 {
-	NODE **aptr;
 	NODE *tmp;
 
 	if (PROCINFO_node == NULL)
 		return;
 	tmp = make_string(subscript, strlen(subscript));
-	aptr = assoc_lookup(PROCINFO_node, tmp);
-	unref(tmp);
-	unref(*aptr);
-	*aptr = make_number(val);
+	assoc_set(PROCINFO_node, tmp, make_number(val));
 }
 
 /* set_FPAT --- handle an assignment to FPAT */
@@ -1427,11 +1436,10 @@ void
 set_FPAT()
 {
 	static bool warned = false;
-	static NODE *save_fpat = NULL;
 	bool remake_re = true;
 	NODE *fpat;
 
-	if (do_lint && ! warned) {
+	if (do_lint_extensions && ! warned) {
 		warned = true;
 		lintwarn(_("`FPAT' is a gawk extension"));
 	}
@@ -1450,9 +1458,9 @@ set_FPAT()
 	 * This comparison can't use cmp_nodes(), which pays attention
 	 * to IGNORECASE, and that's not what we want.
 	 */
-	if (save_fpat
-		&& FPAT_node->var_value->stlen == save_fpat->stlen
-		&& memcmp(FPAT_node->var_value->stptr, save_fpat->stptr, save_fpat->stlen) == 0) {
+	if (save_FPAT
+		&& FPAT_node->var_value->stlen == save_FPAT->stlen
+		&& memcmp(FPAT_node->var_value->stptr, save_FPAT->stptr, save_FPAT->stlen) == 0) {
 		if (FPAT_regexp != NULL)
 			FPAT_regexp = (IGNORECASE ? FPAT_re_no_case : FPAT_re_yes_case);
 
@@ -1465,8 +1473,8 @@ set_FPAT()
 		}
 	}
 
-	unref(save_fpat);
-	save_fpat = dupnode(FPAT_node->var_value);
+	unref(save_FPAT);
+	save_FPAT = dupnode(FPAT_node->var_value);
 	refree(FPAT_re_yes_case);
 	refree(FPAT_re_no_case);
 	FPAT_re_yes_case = FPAT_re_no_case = FPAT_regexp = NULL;
@@ -1559,7 +1567,7 @@ incr_scan(char **scanp, size_t len, mbstate_t *mbs)
  *         field_found = match(substr(string, parse_start), pattern)
  *         
  *         # check for an invalid null field and retry one character away
- *         if (nf > 0 && field_found && RSTART==1 && RLENGTH==0) {
+ *         if (nf > 0 && field_found && RSTART == 1 && RLENGTH == 0) {
  *             parse_start++
  *             field_found = match(substr(string, parse_start), pattern)
  *         }
